@@ -7,7 +7,8 @@ import os.path
 from collections import defaultdict
 from dataclasses import dataclass, field
 from datetime import datetime, timedelta
-from typing import Optional, List, Dict, Iterable, Tuple, Type, Union, Any
+from typing import Optional, List, Dict, Iterable, Tuple, Type, Union, Any, \
+    ClassVar
 from zipfile import ZipFile
 
 import yaml
@@ -701,12 +702,54 @@ class PKIArchitecture:
         'keyset', 'entities', 'certs', 'services', 'entity-defaults'
     )
 
+    @classmethod
+    def build_architecture(cls, arch_label: ArchLabel, cfg: dict,
+                           key_sets: KeySets, external_url_prefix,
+                           extension_plugins: ExtensionPluginRegistry = None,
+                           service_plugins: 'ServicePluginRegistry' = None,
+                           cert_cache=None) -> 'PKIArchitecture':
+        check_config_keys(arch_label, PKIArchitecture.CONFIG_KEYS, cfg)
+        key_set_label = cfg.get('keyset', arch_label)
+        try:
+            key_set = key_sets[key_set_label]
+        except KeyError as e:
+            raise ConfigurationError(
+                f"There is no registered key set with label {key_set_label}"
+            ) from e
+
+        try:
+            entity_cfg = cfg['entities']
+        except KeyError as e:
+            raise ConfigurationError(
+                "The 'entities' key is required in all PKI architecture "
+                "specifications."
+            ) from e
+        try:
+            cert_specs = cfg['certs']
+        except KeyError as e:
+            raise ConfigurationError(
+                "The 'certs' key is required in all PKI architecture "
+                "specifications."
+            ) from e
+        entities = EntityRegistry(
+            entity_cfg, cfg.get('entity-defaults', None)
+        )
+        services = cfg.get('services', {})
+        return PKIArchitecture(
+            arch_label, key_set=key_set, entities=entities,
+            cert_spec_config=cert_specs, service_config=services,
+            external_url_prefix=external_url_prefix,
+            extension_plugins=extension_plugins,
+            service_plugins=service_plugins, cert_cache=cert_cache
+        )
+
     @classmethod
     def build_architectures(cls, key_sets: KeySets, cfgs, external_url_prefix,
                             config_search_dir: Optional[SearchDir],
                             extension_plugins: ExtensionPluginRegistry = None,
                             service_plugins: 'ServicePluginRegistry' = None):
         for arch_label, cfg in cfgs.items():
+            arch_label = ArchLabel(arch_label)
             # external config
             if isinstance(cfg, str):
                 if config_search_dir is None:
@@ -717,58 +760,20 @@ class PKIArchitecture:
                 cfg_path = config_search_dir.resolve(path=cfg)
                 with open(cfg_path, 'r') as external_conf:
                     cfg = yaml.safe_load(external_conf)
-            arch_label = ArchLabel(arch_label)
-            check_config_keys(arch_label, PKIArchitecture.CONFIG_KEYS, cfg)
-            service_base_url = cfg.get(
-                'base-url', f"/{arch_label}"
-            )
-            key_set_label = cfg.get('keyset', arch_label)
-            try:
-                key_set = key_sets[key_set_label]
-            except KeyError as e:
-                raise ConfigurationError(
-                    f"There is no registered key set with label {key_set_label}"
-                ) from e
-
-            try:
-                entity_cfg = cfg['entities']
-            except KeyError as e:
-                raise ConfigurationError(
-                    "The 'entities' key is required in all PKI architecture "
-                    "specifications."
-                ) from e
-            try:
-                cert_specs = cfg['certs']
-            except KeyError as e:
-                raise ConfigurationError(
-                    "The 'certs' key is required in all PKI architecture "
-                    "specifications."
-                ) from e
-            entities = EntityRegistry(
-                entity_cfg, cfg.get('entity-defaults', None)
-            )
-            services = cfg.get('services', {})
-            yield PKIArchitecture(
-                arch_label, key_set=key_set, entities=entities,
-                cert_spec_config=cert_specs,
-                service_config=services,
+            yield cls.build_architecture(
+                arch_label=arch_label, cfg=cfg, key_sets=key_sets,
                 external_url_prefix=external_url_prefix,
-                service_base_url=service_base_url,
                 extension_plugins=extension_plugins,
                 service_plugins=service_plugins
             )
 
     def __init__(self, arch_label: ArchLabel,
                  key_set: KeySet, entities: EntityRegistry,
-                 cert_spec_config, service_config,
-                 external_url_prefix, service_base_url,
+                 cert_spec_config, service_config, external_url_prefix,
                  extension_plugins: ExtensionPluginRegistry = None,
-                 service_plugins: 'ServicePluginRegistry' = None):
+                 service_plugins: 'ServicePluginRegistry' = None,
+                 cert_cache=None):
 
-        if not service_base_url.startswith('/'):
-            raise ConfigurationError(
-                "Service base URL should start with '/'."
-            )
         self.arch_label = arch_label
         self.key_set = key_set
         self.entities = entities
@@ -777,7 +782,7 @@ class PKIArchitecture:
             extension_plugins or DEFAULT_EXT_PLUGIN_REGISTRY
 
         self.service_registry: ServiceRegistry = ServiceRegistry(
-            self, external_url_prefix, service_base_url, service_config,
+            self, external_url_prefix, service_config,
             plugins=service_plugins
         )
 
@@ -839,7 +844,7 @@ class PKIArchitecture:
             self._cert_labels_by_issuer[spec.issuer].append(name)
             self._cert_labels_by_subject[spec.subject].append(name)
 
-        self._cert_cache = {}
+        self._cert_cache = cert_cache if cert_cache is not None else {}
 
     def get_cert_spec(self, label: CertLabel) -> CertificateSpec:
         try:
@@ -974,7 +979,6 @@ class PKIArchitecture:
 
     def dump_certs(self, folder_path: str, use_pem=True, flat=False,
                    include_pkcs12=False):
-        self._load_all_certs()
         os.makedirs(folder_path, exist_ok=True)
         itr = self._dump_certs(
             use_pem=use_pem, flat=flat, include_pkcs12=include_pkcs12
@@ -1125,6 +1129,10 @@ class PKIArchitecture:
 @dataclass(frozen=True)
 class ServiceInfo(ConfigurableMixin):
     """Base class to describe a PKI service."""
+
+    arch_label: ArchLabel
+    """Architecture to which the service belongs. """
+
     label: ServiceLabel
     """
     Label by which the service is referred to within Certomancer configuration.
@@ -1135,16 +1143,13 @@ class ServiceInfo(ConfigurableMixin):
     Prefix that needs to be prepended to produce a "fully qualified" URL.
     """
 
-    base_url: str
-    """
-    Base URL from which the service's "siblings" branch off.
-    This excludes the service label and external URL prefix.
-    """
+    base_url: ClassVar[str]
 
     @property
     def internal_url(self) -> str:
         """
-        Internal URL for the service, i.e. without the external URL prefix.
+        Internal URL for the service, i.e. without the external URL prefix
+        or the arch_label prefix
         """
 
         return f"{self.base_url}/{self.label}"
@@ -1154,13 +1159,16 @@ class ServiceInfo(ConfigurableMixin):
         """
         Full URL where the service's main endpoint can be found.
         """
-        return f"{self.external_url_prefix}{self.base_url}/{self.label}"
+        return \
+            f"{self.external_url_prefix}/{self.arch_label}{self.internal_url}"
 
 
 @dataclass(frozen=True)
 class OCSPResponderServiceInfo(ServiceInfo):
     """Configuration describing an OCSP responder."""
 
+    base_url = '/ocsp'
+
     for_issuer: EntityLabel
     """
     Issuing entity on behalf of which the responder acts.
@@ -1231,6 +1239,8 @@ class OCSPResponderServiceInfo(ServiceInfo):
 class TSAServiceInfo(ServiceInfo):
     """Configuration describing a time stamping service."""
 
+    base_url = '/tsa'
+
     signing_cert: CertLabel
     """
     Label of the signer's certificate.
@@ -1293,6 +1303,8 @@ class CRLRepoServiceInfo(ServiceInfo):
         parameter.
     """
 
+    base_url = '/crls'
+
     for_issuer: EntityLabel
     """
     Issuing entity for which the CRLs are issued.
@@ -1373,6 +1385,7 @@ class CRLRepoServiceInfo(ServiceInfo):
 
 @dataclass(frozen=True)
 class CertRepoServiceInfo(ServiceInfo):
+    base_url = '/certs'
     for_issuer: EntityLabel
     issuer_cert: Optional[CertLabel] = None
     publish_issued_certs: bool = True
@@ -1394,6 +1407,11 @@ class CertRepoServiceInfo(ServiceInfo):
         fname = CertRepoServiceInfo.issuer_cert_file_name(use_pem=use_pem)
         return f"{self.url}/{fname}"
 
+    @property
+    def issuer_cert_internal_url(self):
+        fname = CertRepoServiceInfo.issuer_cert_file_name(use_pem=True)
+        return f"{self.internal_url}/{fname}"
+
     def issued_cert_url(self, label: CertLabel, use_pem=True):
         path = self.issued_cert_url_path(label=label, use_pem=use_pem)
         return f"{self.url}/{path}"
@@ -1405,6 +1423,8 @@ class PluginServiceInfo(ServiceInfo):
     Configuration describing a service provided by a service plugin.
     """
 
+    base_url = '/plugin'
+
     plugin_label: PluginLabel
     """
     Label of the service plugin.
@@ -1420,6 +1440,15 @@ class PluginServiceInfo(ServiceInfo):
     The content type of the response returned by the plugin.
     """
 
+    @property
+    def internal_url(self) -> str:
+        """
+        Internal URL for the service, i.e. without the external URL prefix
+        or the arch_label prefix
+        """
+
+        return f"{self.base_url}/{self.plugin_label}/{self.label}"
+
 
 class OCSPInterface(RevocationInfoInterface):
 
@@ -1603,18 +1632,16 @@ class ServiceRegistry:
     """
 
     def __init__(self, pki_arch: PKIArchitecture, external_url_prefix,
-                 base_url, service_config,
-                 plugins: ServicePluginRegistry = None):
-        self.services_base_url = base_url
+                 service_config, plugins: ServicePluginRegistry = None):
         self.pki_arch = pki_arch
         self.plugins = plugins or DEFAULT_SRV_PLUGIN_REGISTRY
 
-        def _gen_svc_config(url_suffix, configs):
+        def _gen_svc_config(configs):
             for lbl, cfg in configs.items():
                 cfg = dict(cfg)
                 cfg.setdefault('external-url-prefix', external_url_prefix)
-                cfg.setdefault('base-url', f"{base_url}/{url_suffix}")
                 cfg['label'] = lbl
+                cfg['arch_label'] = pki_arch.arch_label.value
                 yield ServiceLabel(lbl), cfg
 
         check_config_keys(
@@ -1627,22 +1654,22 @@ class ServiceRegistry:
         self._ocsp = {
             label: OCSPResponderServiceInfo.from_config(cfg)
             for label, cfg
-            in _gen_svc_config('ocsp', service_config.get('ocsp', {}))
+            in _gen_svc_config(service_config.get('ocsp', {}))
         }
         self._crl_repo = {
             label: CRLRepoServiceInfo.from_config(cfg)
             for label, cfg
-            in _gen_svc_config('crls', service_config.get('crl-repo', {}))
+            in _gen_svc_config(service_config.get('crl-repo', {}))
         }
         self._cert_repo = {
             label: CertRepoServiceInfo.from_config(cfg)
             for label, cfg
-            in _gen_svc_config('certs', service_config.get('cert-repo', {}))
+            in _gen_svc_config(service_config.get('cert-repo', {}))
         }
         self._tsa = {
             label: TSAServiceInfo.from_config(cfg)
             for label, cfg
-            in _gen_svc_config('tsa', service_config.get('time-stamping', {}))
+            in _gen_svc_config(service_config.get('time-stamping', {}))
         }
 
         plugin_cfg = service_config.get('plugin', {})
@@ -1652,16 +1679,14 @@ class ServiceRegistry:
         def _cfg_plugin(plugin_label, cfg_for_plugin):
             plugin = self.plugins[plugin_label]
             content_type = plugin.content_type
-            svc_configs = _gen_svc_config(
-                f'plugin/{plugin_label}', cfg_for_plugin
-            )
+            svc_configs = _gen_svc_config(cfg_for_plugin)
             for service_label, cfg in svc_configs:
                 yield service_label, PluginServiceInfo(
                     plugin_label=plugin_label, content_type=content_type,
                     plugin_config=plugin.process_plugin_config(cfg),
                     label=service_label,
                     external_url_prefix=cfg['external-url-prefix'],
-                    base_url=cfg['base-url'],
+                    arch_label=pki_arch.arch_label
                 )
 
         self._plugin_services = {
@@ -1970,4 +1995,9 @@ class CertomancerConfig:
         }
 
     def get_pki_arch(self, label: ArchLabel) -> PKIArchitecture:
-        return self.pki_archs[label]
+        try:
+            return self.pki_archs[label]
+        except KeyError as e:
+            raise ConfigurationError(
+                f"There is no PKI architecture with label {label}."
+            ) from e
